import 'package:flutter/widgets.dart';
import 'package:sunny_dart/sunny_dart.dart';
import 'package:flexidate/flexidate.dart';
import 'package:sunny_sdk_core/api_exports.dart';
import 'package:sunny_sdk_core/model_exports.dart';
import 'package:sunny_sdk_core/sunny_sdk_core.dart';
import 'package:sunny_service_stubs/contact.dart';
import 'package:sunny_service_stubs/models/facts/facts.dart';
import 'package:sunny_service_stubs/models/fact_schemas/fact_schemas.dart';
import 'package:sunny_service_stubs/services/named_icons.dart';

import 'package:timezone/timezone.dart';

import '../facts.dart';

mixin BaseFactMixin {}

mixin BaseFactImpl implements IFact {
  IFactSchema? _factSchema;

  IFactSchema get factSchema => this.factSchemaOrNull ?? nullPointer("Missing fact schema");
  IFactSchema? get factSchemaOrNull => _factSchema ??= factService.getFactSchema(this.mtype);

  set factSchema(IFactSchema factSchema) {
    _factSchema = factSchema;
  }

  MKey? get contactKey;

  set contactKey(MKey? key);

  MKey? get author => null;

  Widget? icon(IRef contact, {bool? circular, double? size = kTileIconSize}) {
    return NamedIcons.getIconOrNull(factSchema.icon)?.copyWith(circular: circular, isSolid: circular, size: size).icon;
  }

  bool involves(MKey contactId) {
    IFact self = this;
    return self is MultiParticipantFact ? (self as MultiParticipantFact).involvesAny(contactId) : contactKey == contactId;
  }

  // Widget titleTrailing(BuildContext context, Contact contact) => null;

  TZDateTime? getDate(IFactMetaDate metaDate, Location location) {
    final date = getByPath(metaDate.path!);
    if (date == null) return null;
    TZDateTime? resolvedDate;

    if (date is TZDateTime) {
      resolvedDate = date;
    } else if (date is DateTime) {
      resolvedDate = TZDateTime.from(date, location);
    } else if (date is FlexiDate) {
      final now = DateTime.now();
      var stepDate = cloneDate(now, year: date.year, month: date.month, day: date.day);
      while (stepDate.isPast) {
        stepDate = cloneDate(stepDate, year: stepDate.year + 1);
      }
      resolvedDate = TZDateTime.from(stepDate, location);
    }
    return resolvedDate;
  }

  FlexiDate? getFlexiDate(IFactMetaDate metaDate) {
    final date = getByPath(metaDate.path!);
    if (date is FlexiDate) {
      return date;
    }
    return null;
  }

  bool get showTimestamp => factSchemaOrNull?.isHistorical == true;

  // @override
  // Widget buildTile(BuildContext context,
  //     {@required Contact contact,
  //     bool isCard = true,
  //     VoidCallback onTap,
  //     VoidCallback changed,
  //     Widget trailing}) {
  //   return FactCell(
  //       key: Key("fact-cell-${this.id}"),
  //       isCard: isCard,
  //       showTimestamp: showTimestamp,
  //       timestamp: smartNoteDate,
  //       child: PlatformListTile(
  //         dense: false,
  //         card: isCard,
  //         onTap: onTap ??
  //             () async {
  //               final schema = await factService.loadFactSchema(this.mtype);
  //               await openFactEditPage(
  //                 context,
  //                 contact: contact,
  //                 fact: this as Fact,
  //                 factSchema: schema,
  //               );
  //             },
  //         titleTrailing: titleTrailing(context, contact),
  //         leading: icon(contact, size: tileIconSize(dense: isCard == false)),
  //         trailing: trailing,
  //         minHeight: 70,
  //         title: TileTitle(title(contact)),
  //         subtitle: TileSubtitle(subtitle(contact)),
  //       ));
  // }

  // Widget buildTileForContact(
  //   BuildContext context, {
  //   Key key,
  //   @required Contact contact,
  //   VoidCallback changed,
  //   Widget trailing,
  //   VoidCallback onTap,
  // }) {
  //   return FactCell(
  //       showTimestamp: showTimestamp,
  //       timestamp: smartNoteDate,
  //       child: PlatformListTile(
  //         dense: false,
  //         onTap: onTap,
  //         leading:
  //             contact == null ? icon(contact) : ContactAvatar(contact: contact),
  //         trailing: trailing,
  //         title: TileTitle(contact?.fullName?.join(title(contact), " - ")),
  //         subtitle: TileSubtitle(subtitle(contact)),
  //       ));
  // }

  @override
  DateTime? get smartNoteDate {
    final firstHistorical = _factSchema?.firstHistoricalDate;
    if (firstHistorical == null) {
      return null;
    } else {
      if (firstHistorical.isFlexible == true) {
        FlexiDate? flexiDate = this.getByPath(firstHistorical.path!);
        if (flexiDate?.isFullDate == true) {
          return flexiDate?.toDateTime();
        } else {
          /// If the user doesn't have a year set, put it at the top of the timeline
          return this.dateCreated ?? DateTime.now();
        }
      } else {
        return this.getByPath(firstHistorical.path!);
      }
    }
  }

  // String shortTitle(Contact contact) => title(contact);

  bool get isAttribute => factSchemaOrNull?.isAttribute == true;

  bool get isPending => factSchemaOrNull?.isActionable == true
      && factSchemaOrNull?.isTimeSensitive == true && isFuture(smartNoteDate);

  bool get isHistory => factSchemaOrNull?.isActionable == true &&
      factSchemaOrNull?.isTimeSensitive == true &&
      isPast(smartNoteDate);

  initialize(MKey targetKey, MKey agentKey) {
    dateCreated = DateTime.now();
    contactKey = targetKey;
  }
}

extension FactIterableExtension on Iterable<Fact> {
  Iterable<Fact> findAll(MSchemaRef type) => this.where((f) => f.mmeta.mtype == type);

  Fact? find(MSchemaRef type) => this.firstWhere((f) => f.mmeta.mtype == type);

  Fact get(MSchemaRef type) => this.firstWhere((f) => f.mmeta.mtype == type);
}
