// ignore_for_file: unused_import, override_on_non_overriding_member
import 'dart:async';

import 'package:dartxx/dartxx.dart';
import 'package:facts/model/assisted_task.dart';
import 'package:facts/model/birthday.dart';
import 'package:flutter/material.dart';
import 'package:sunny_dart/sunny_dart.dart';
import 'package:sunny_sdk_core/model_exports.dart';
import 'package:sunny_service_stubs/models/fact_schemas/fact_schemas.dart';
import 'package:sunny_service_stubs/sunny_service_stubs.dart';

import '../facts.dart';
import 'fact_mixin.dart';

export 'package:sunny_service_stubs/meta/smart_notes_item_mixin.dart';

export "fact_mixin.dart";
export "membership_fact_mixins.dart";
export 'relive_it_mixin.dart';

mixin BirthdayMixin implements RemindableFact, IBirthday {
  @override
  String title(IRef contact) {
    var title = "";
    // if (this.contactKey?.mxid != contact.id) {
    //   title += "${contactService.load(this.contactKey!).resolveOrNull()?.fullName ?? 'Unnamed'}\n";
    // }
    title += "Birthday on ${birthday!.formatted()}";
    return title;
  }

  @override
  String remindLabel(IFactMetaDate metaDate) {
    return "Birthday";
  }

  @override
  FlexiDate? getPrimaryDate() => birthday;

  @override
  String? subtitle(IRef contact) {
    if (birthday!.isFullDate) {
      final birthdayThisYear = birthday!.toDateTime();
      final age = birthdayThisYear.yearsApart.abs();
      final verb = birthdayThisYear.isPast ? "Turned" : "Turns";
      if (age > 40) {
        return "$verb $age years young this year";
      } else {
        return "$verb $age this year";
      }
    }
    return null;
  }
}

mixin ContactInterestsMixin implements IContactInterests {
  String title(IRef contact) {
    return illegalState();
    // final list =
    //     interestService.getAll(this.interests).joinWithAnd((_) => _.title);
    // return "Interested in $list";
  }

  String? subtitle(IRef contact) {
    return null;
  }

// Widget buildTile(BuildContext context,
//     {@required Contact contact, bool isCard = true, VoidCallback onTap, VoidCallback changed, Widget trailing}) {
//   final groupedInterests = interestService.getAll(this.interests).groupBy((interest) {
//     return interest.parent?.key;
//   });
//   final groups = interestService.getAll(groupedInterests.keys).keyed((group) => group.key);
//   return column(
//     [
//       for (final group in groups.values)
//         PlatformListTile(
//           leading: group.icon?.solid()?.widget ?? Icon(SunnyIcons.interest.solid(), size: 22),
//           title: group.title == null ? null : Text(group.title),
//           onTap: onTap,
//           subtitle: Wrap(
//             spacing: 5,
//             runSpacing: 5,
//             children: [
//               for (final interest in groupedInterests[group.key])
//                 InputChip(
//                   key: ObjectKey(interest),
//                   onSelected: (b) {},
//                   labelPadding: interest.icon == null ? null : SunnyMargins.right,
//                   padding: SunnyMargins.zero,
//                   label: Text(interest.title),
//                   visualDensity: VisualDensity.compact,
//                   avatar: interest.icon
//                       ?.solid(SunnyColors.primaryAction)
//                       ?.copyWith(foregroundColor: SunnyColors.white, size: 24)
//                       ?.widget,
//                   materialTapTargetSize: MaterialTapTargetSize.shrinkWrap,
//                 ),
//             ],
//           ),
//         ),
//     ],
//   );
// }
}

// OperationInfo defaultOperation(MOperationRef opRef) => OperationInfo(
//       opRef,
//       Icons.check_box,
//       MOperationDefinition.of(
//         operationRef: opRef,
//         isAssisted: true,
//         isVirtual: true,
//         operationLabel: "Loading...",
//       ),
//     );

mixin AssistedTaskMixin implements IAssistedTask {
  // set contactKey(MKey contactKey) {
  //   this["contactKey"] = contactKey;
  // }
  //
  // MKey get contactKey => jpath(AssistedTaskPaths.contactKey);
  //
  bool get isSkippable => true;

  // FutureOr<OperationInfo> get operationInfo => operationOptionsHandler
  //     .thenOr((handler) => handler.baseHandler.loadValue(operationRef))
  //     .unbox();
  //
  // OperationInfo get operationInfoOrDefault =>
  //     operationInfo.resolve(defaultOperation(operationRef));

  // Future<IRef?> get contact async {
  //   final contactKey = this.contactKey;
  //   if (contactKey == null) return null;
  //   return await contactService.load(contactKey);
  // }

  bool get isDue => (displayDate ?? dueDate ?? DateTime.now()).isPast;

  // @override
  // Widget icon(Contact contact, {bool circular, double size = kTileIconSize}) {
  //   return setup?.asFact()?.icon(contact, circular: circular, size: size) ??
  //       (operationIcons[operationRef] ??
  //               NamedIcons.getIconOrNull(factSchema.icon))
  //           .copyWith(circular: circular, size: size, isSolid: circular == true)
  //           .icon;
  // }

  // SunnyPackAction toSunnyPackAction() =>
  //     SunnyPackAction.ofTask(task: this as AssistedTask);

  @override
  String title(IRef contact) {
    final t = setup?.asFact()?.title(contact) ?? taskTitle?.nullIfBlank() ?? taskTitle ?? "Unnamed Task";
    return t;
  }

  // String pageTitle(IRef contact) {
  //   final operationInfo = this.operationInfoOrDefault;
  //   return setup?.asFact()?.title(contact) ??
  //       taskTitle?.nullIfBlank() ??
  //       operationInfo?.label?.recase?.titleCase ??
  //       taskTitle ??
  //       "Unnamed Task";
  // }

  String get formattedDueLabel {
    String dueLabel = "";
    bool formatTime = dueDate?.hasTime ?? true;
    final dueConfig = this.dueConfig as TaskDueConfig?;
    if (dueConfig is MetaDateDueConfig) {
      formatTime = factService.getFactSchema(dueConfig.schemaRef)?.firstHistoricalDate?.hasTime == true;
    } else if (dueConfig is SmartDateDueConfig) {
      formatTime = dueConfig.hasTime ?? true;
    }
    final String formatted = [
      dueConfig?.dueConfigLabel,
      smartDateFormat(isCompleted ? completionDate : dueDate, formatTime: formatTime)
    ].where((s) => s.isNotNullOrBlank).join(": ");
    dueLabel += formatted;
    return dueLabel;
  }

  subtitle(IRef contact) {
    final lines = <String?>[];
    // final operationInfo = this.operationInfoOrDefault;
    lines.addAll(taskDescription.toStringList());
    lines.add(setup?.asFact()?.subtitle(contact));

    String subtitle = "";
    subtitle += isCompleted ? "Completed: " : "Due: ";
    subtitle += formattedDueLabel;

    lines.add(subtitle);
    return lines.where((str) => str.isNotNullOrBlank).join("\n");
  }

  bool get isCompleted => completionDate != null;

  void initialize(MKey targetKey, MKey agentKey) {
    dateCreated = DateTime.now();
    isUnread = false;
    contactKey = targetKey;
    assigneeKey = agentKey;
  }

  get diffKey => mkey?.mxid;

  get diffSource => {"id": mkey?.mxid, "mmodified": (this as MEntity).mmeta.mmodified};
}

abstract class IDueConfig {
  String? get dueConfigEmoji;

  String? get dueConfigLabel;
}

mixin TaskDueConfigMixin implements IDueConfig {
  String? get dueConfigEmoji => 'smiley';

  String? get dueConfigLabel => null;
}

mixin TaskFactMixin implements ITaskFact {
  // title(IRef contact) => this.taskTitle;

  // subtitle(IRef contact) => this.taskDescription;

  bool get isHistorical => completionDate != null;

  bool get isComplete => completionDate != null;

  DateTime? get smartNoteDate => isHistorical ? completionDate : dueDate;

  DateTime? get dueOrDisplayDate {
    return displayDate ?? dueDate;
  }
}

mixin FactsMixin implements IFacts {
  T? getFact<T extends Fact>() => data!.whereType<T>().firstOr();
}
mixin EmployerMixin implements IEmployer {
  String? title(IRef contact) {
    return name;
  }

  String? subtitle(IRef contact) => null;
}

mixin EmployeeMixin implements IEmployee, RemindableFact {
  String title(IRef contact) {
    String title = "";
    if (this.endDate != null) {
      title += "Worked ";
    } else {
      title += "Works ";
    }
//    if (this.isCoworker == true) {
//      title += "with you ";
//    }
//    if (employer?.name != null) {
//      title += "at ${employer.name} ";
//    }
    if (this.company?.name?.isNotNullOrBlank == true) {
      title += "at ${company!.name} ";
    }
    if (position != null) {
      title += "as $position";
    }

    return title;
  }

  @override
  FlexiDate? getPrimaryDate() => startDate ?? endDate;

  String subtitle(IRef contact) {
    String subtitle = "";
    if (startDate != null) {
      subtitle += endDate == null ? "since " : "from ";
      subtitle += "${startDate!.formatted()} ";
    }
    if (endDate != null) {
      subtitle += "until ${endDate!.formatted()}";
    }

    return subtitle;
  }

  String shortTitle(IRef contact) {
    return "started job: ${companyName ?? ''}";
  }

  String? get companyName {
    return company?.name;
  }

  String? remindLabel(IFactMetaDate metaDate) {
    final companyNameString = companyName == null ? "their job" : "at $companyName";
    if (metaDate.path == EmployeePaths.startDate) {
      return "starting $companyNameString";
    } else if (metaDate.path == EmployeePaths.endDate) {
      return "leaving $companyNameString";
    }
    return null;
  }
}

mixin CongregantMixin implements ICongregant, RemindableFact {
  String title(IRef contact) {
    String title = "";
//    if (this.endDate != null) {
//      title += "Attended ${church?.name ?? "church"}";
//    } else {
    title += "Attends ${church?.name ?? "church"}";
//    }

    return title;
  }

  FlexiDate? getPrimaryDate() {
    return null;
  }

  String subtitle(IRef contact) {
    String subtitle = "";
    if (churchPosition != null) {
      subtitle += "$churchPosition ";
    }
//    if (startDate != null) {
//      subtitle += endDate == null ? "since " : "from ";
//      subtitle += "${startDate.formatted()} ";
//    }
//    if (endDate != null) {
//      subtitle += "until ${endDate.formatted()}";
//    }

    return subtitle;
  }

  String shortTitle(IRef contact) {
    return "joined church";
  }

  String? remindLabel(IFactMetaDate metaDate) {
//    if (metaDate.path == CongregantPaths.startDate) {
//      return "joined a church";
//    } else if (metaDate.path == CongregantPaths.endDate) {
//      return "left their church";
//    }
    return null;
  }
}

mixin FamilyMixin implements IFamily {
  MKey? contactKey;

  String title(IRef contact) {
    String title = "Family";
    return title;
  }

  String? subtitle(IRef contact) => null;

  String shortTitle(IRef contact) => "Family";

  String remindLabel(_) {
    return "Family";
  }
}

mixin PetMixin implements IPet, RemindableFact {
  String title(IRef contact) {
    String title = "Pet ${petType == null ? "" : "$petType: "}$name";
    return title;
  }

  String subtitle(IRef contact) => birthday!.fullFormat(historyLabel: "old");

  String shortTitle(IRef contact) => "${petType == null ? null : "$petType: "}$name";

  String remindLabel(IFactMetaDate metaDate) {
    return name != null ? "$name's b-day" : "pet's bday";
  }

  @override
  FlexiDate? getPrimaryDate() => birthday;
}

mixin MemoryMixin implements IMemory, HasImage, HasImages, RemindableFact, LabeledFact, MultiParticipantFact {
  List<IImageContent> get imageList => photos!;

  MKey? get contactKey => contactKeys?.firstOrNull();

  IImageContent get imageContent => imageList.first;

  set contactKey(MKey? contactKey) {
    this.contactKeys = {
      ...?contactKeys,
      if (contactKey != null) contactKey,
    }.toList();
  }

  MKey? get author {
    /// Find the one not linked to an org... hacky
    return contactKeys.orEmpty().where((k) {
      return k.mtype.artifactId?.toLowerCase().contains("contact") == true;
    }).firstOr();
  }

  @override
  String? title(IRef contact) {
    String title = memory ?? 'Memory';
    return title;
  }

  @override
  String? shortTitle(IRef contact) => details;

  @override
  String? subtitle(IRef contact) {
    return joinString((_) {
      // _ += memoryDate.fullFormat(withYear: true);
      _ += details;
    }, '\n');
  }

  String remindLabel(IFactMetaDate metaDate) {
    return "Memory";
  }

  @override
  FlexiDate? getPrimaryDate() => memoryDate;

  void initialize(MKey targetKey, MKey agentKey) {
    initialize(targetKey, agentKey);
    this.possessions ??= [];
    this.pets ??= [];
    this.traditions ??= [];
    this.traits ??= [];
    this.memoryDate ??= FlexiDateData.now();
  }
}

mixin MilestoneMixin on FactMixin implements IMilestone {
  MKey? get contactKey => contactKeys!.firstOrNull();

  set contactKey(MKey? contactKey) {
    this.contactKeys = {
      ...?contactKeys,
      if (contactKey != null) contactKey,
    }.toList();
  }

  String title(IRef contact) {
    String title = "Milestone: ${milestone ?? 'No Title'}";
    return title;
  }

  String? shortTitle(IRef contact) => details;

  String subtitle(IRef contact) {
    String history = milestoneDate!.fullFormat(withYear: true);
    if (details != null) history += "\n$details";
    return history;
  }

  void initialize(MKey targetKey, MKey agentKey) {
    super.initialize(targetKey, agentKey);
    milestoneDate ??= FlexiDateData.now();
  }
}

mixin MetaDateDueConfigMixin implements IMetaDateDueConfig {
  String get dueConfigEmoji => factService.getFactEmoji(schemaRef);

  String? get dueConfigLabel => factService.getFactSchema(schemaRef)?.label?.capitalize();
}

mixin GiftMixin implements IGift {
  MKey? get contactKey => this.gifteeKeys!.firstOrNull();

  set contactKey(MKey? contactKey) {
    this.gifteeKeys = {
      ...?gifteeKeys,
      if (contactKey != null) contactKey,
    }.toList();
  }

  @override
  String? title(IRef contact) {
    String title = "Gave gift";
    if (gift.isNotNullOrBlank) title += ": $gift";
    return title;
  }

  @override
  String? subtitle(IRef contact) {
    String title = "Occasion / Date:  ";
    if (occasion.isNotNullOrBlank) {
      title += "$occasion ";
    }
    return title;
  }

  void initialize(MKey targetKey, MKey agentKey) {
    dateCreated = DateTime.now();
    gifteeKeys = [agentKey];
    recipientKey = targetKey;
  }
}

mixin NoteMixin implements INote {
  String title(IRef contact) => noteTitle ?? "Added a Note";

  String? subtitle(IRef contact) {
    return nonBlank(notes);
  }
}

mixin OwesMoneyMixin implements IOwesMoney {
  void initialize(MKey targetKey, MKey agentKey) {
    dateCreated = DateTime.now();
    creditor = agentKey;
    debtor = targetKey;
  }

  set contactKey(MKey? mkey) {
    if (this.creditor == null && this.debtor == null) {
      this.creditor = mkey;
    }
  }

  MKey? get contactKey => creditor;

  String title(IRef contact) {
    final linkType = debtor.linkType(contact);
    String title = "";
    switch (linkType) {
      case ContactLinkType.contact:
        title += "Owes you ";
        break;
      case ContactLinkType.user:
        title += "Pay ";
        break;
      default:
        title += "Owed ";
        break;
    }
    if (this.amount != null) {
      title += "\$${this.amount!.toStringAsFixed(2)} ";
    } else {
      title += "money";
    }
    return title;
  }

  String? subtitle(IRef contact) {
    String subtitle = "";
    if (this.reason != null) {
      subtitle += "$reason ";
    }
    return nonBlank(subtitle.trim());
  }

  bool isContactDebtor(IRef contact) {
    return debtor?.mxid == contact.id;
  }

  bool isContactCreditor(IRef contact) {
    return creditor?.mxid == contact.id;
  }

// Widget icon(Contact contact, {bool circular, double size = kTileIconSize}) {
//   if (isContactCreditor(contact)) {
//     return NamedIcons.getPaid
//         .copyWith(size: size, circular: circular, isSolid: circular)
//         .widget;
//   } else if (isContactDebtor(contact)) {
//     return NamedIcons.pay
//         .copyWith(size: size, circular: circular, isSolid: circular)
//         .widget;
//   } else {
//     return NamedIcons.money
//         .copyWith(size: size, circular: circular, isSolid: circular)
//         .widget;
//   }
// }
}

mixin PaymentMixin implements IPayment {
  set contactKey(MKey? contactKey) {
    this.payor = contactKey;
  }

  MKey? get contactKey => payor;

  void initialize(MKey targetKey, MKey agentKey) {
    dateCreated = DateTime.now();
    paymentDate = DateTime.now();
    payor = agentKey;
    payee = targetKey;
  }

  String title(IRef contact) {
    // Determining if the payor if you or someone else
    final linkType = payor.linkType(contact);
    String title = "";
    switch (linkType) {
      // You are the one who was PAID
      case ContactLinkType.contact:
        title += "Paid You ";
        break;
      // You are the one who PAID
      case ContactLinkType.user:
        title += "You Paid ";
        break;
      default:
        title += "Paid ";
        break;
    }

    if (this.amount != null) {
      title += "\$${this.amount!.toStringAsFixed(2)} ";
    } else {
      title += "money";
    }
    return title;
  }

  String? subtitle(IRef contact) {
    String subtitle = "";
    if (this.reason != null) {
      subtitle += "$reason ";
    }
    return nonBlank(subtitle.trim());
  }
}

//mixin ReconnectMixin implements IReconnect{
//  String title(IRef contact) {
//    return "Need to reconnect";
//  }
//
//  String subtitle(IRef contact) => ideas?.isNotEmpty == true ? "Notes: $ideas" : null;
//}
mixin SmartDateDueConfigMixin implements ISmartDateDueConfig {
  String? get dueConfigEmoji => null;

  String? get dueConfigLabel => null;
}
mixin ScriptExpressionMixin {}
mixin SingleValueMixin {}
mixin SmsContentMixin {}
mixin StartMixin {}
mixin StartThreadRequestMixin {}
mixin SunnyBundleMixin {}
mixin SunnyBundlesMixin {}
mixin SunnyGroupMixin {}

mixin SunnyPublisherMixin {}
mixin SunnyPublishersMixin {}
mixin TemplateDataMixin {}
mixin TemplateSetMixin {}
mixin TemplateVariationsMixin {}
mixin TextContentMixin {}
mixin UpdateThreadStatusRequestMixin {}
mixin VideoContentMixin {}
mixin WaitPriorMixin {}
mixin YoutubeContentMixin {}
mixin MSchemaDefinitionMixin {}

mixin AnniversaryApiMixin {}
mixin AssistedTasksApiMixin {}
mixin AssistedTasksMixin implements MModelList<IAssistedTask> {}
mixin AssistedTasksListMixin implements MModelList<IAssistedTask> {}

mixin AssistedTaskApiMixin implements IAssistedTaskApi {
  String? keyToId(MKey? key) => key?.recordKey.value;
}

mixin FactAptiMixin {
  String? keyToId(MKey? key) => key?.value;
}

enum ContactLinkType { contact, user, other }

extension MModelFactExt on MModel {
  Fact? asFact() {
    if (this is Fact) return (this as Fact);
    return null;
  }
}
