import 'package:facts/facts.dart';
import 'package:sunny_sdk_core/model_exports.dart';
import 'package:sunny_service_stubs/contact.dart';
import 'package:sunny_service_stubs/models/fact_schemas/model/fact_meta_date.dart';

mixin BookClubMemberMixin implements IBookClubMember {
  String title(IRef contact) {
    return "Member of ${bookClub?.name ?? 'a book club'}";
  }

  String? subtitle(IRef contact) {
    return null;
  }
}

mixin PlayerMixin implements IPlayer {
  String title(IRef contact) {
    return "Played on ${team?.name ?? 'a team'}";
  }

  String? subtitle(IRef contact) {
    return null;
  }
}

mixin HometownMixin implements IHometown {
  String title(IRef contact) {
    return "Hometown: ${this.location1!.locality}, ${this.location1!.region}";
  }

  String? subtitle(IRef contact) {
    return null;
  }
}

mixin ContactExtendedPropsMixin implements IContactExtendedProps {
  String? title(IRef contact) {
//    if (maidenName == null && nickName == null && gender != null) {
    if (nickName == null && gender != null) {
      return gender.toString();
    }
//    if (maidenName != null) return "Maiden Name: $maidenName";
    if (nickName != null) return "Nickname: $nickName";
    return null;
  }

  String? subtitle(IRef contact) {
//    if (maidenName != null && nickName != null) {
    if (nickName != null) {
      return "Nickname: $nickName";
    }
    return null;
  }

  // Widget icon(IRef contact, {bool circular, double size = kTileIconSize}) {
  //   NamedIcon icon;
  //   if (gender != null) {
  //     if (gender == Gender.male_) {
  //       icon = NamedIcons.male;
  //     } else if (gender == Gender.female_) {
  //       icon = NamedIcons.female;
  //     } else {
  //       icon = NamedIcons.otherGender;
  //     }
  //   }
  //   icon ??= NamedIcons.contact;
  //   return icon.copyWith(circular: circular, size: size).widget;
  // }
}

mixin RememberMixin implements IRemember {
  String title(IRef contact) {
    return "Remember: $rememberTitle";
  }

  String? subtitle(IRef contact) {
    return rememberDetails;
  }
}

mixin CollegeStudentMixin implements ICollegeStudent, RemindableFact {
  String title(IRef contact) {
    var str = "Attended";
    return "$str ${university?.name ?? university?.shortName ?? "university"}";
  }

  @override
  String remindLabel(IFactMetaDate metaDate) {
    return "Graduated";
  }

  String? subtitle(IRef contact) {
    String? subtitle;
    if (focus != null) {
      subtitle = "Studied: $focus";
    }
    return subtitle;
  }

  @override
  FlexiDate? getPrimaryDate() => this.startDate ?? this.endDate;
}

mixin PrepSchoolStudentMixin implements IPrepSchoolStudent {
  String title(IRef contact) {
    return "Attended ${school?.name}";
  }

  String remindLabel(IFactMetaDate metaDate) {
    return "Graduated";
  }

  String? subtitle(IRef contact) {
    return graduatingClass != null ? "Graduated $graduatingClass" : null;
  }
}
